/*****************************************************************************\
* (c) Copyright 2000-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#define CALODET_DESUBSUBCALORIMETER_CPP 1
// ============================================================================

// CaloDet
#include "CaloDet/DeSubSubCalorimeter.h"
#include "CaloKernel/CaloException.h"

#include <algorithm> // for std::find

/** @file DeSubSubCalorimeter.cpp
 *
 *  Implementation of class :  DeSubSubCalorimeter
 *  @author Olivier Callot Olivier.Callot@cern.ch
 *  @author Vanya Belyaev  Ivan.Belyaev@itep.ru
 */

// ============================================================================
// Standard Constructors
// ============================================================================
DeSubSubCalorimeter::DeSubSubCalorimeter( const std::string& name )
  : DetectorElement     ( name )
    , m_cellSize( 0.0 )
    , m_xSize(0.0)
    , m_ySize(0.0)
    , m_area( -1  )
{}

// ============================================================================
// Destructor
// ============================================================================
DeSubSubCalorimeter::~DeSubSubCalorimeter(){}


// ============================================================================
/// object identification
// ============================================================================
const CLID& DeSubSubCalorimeter::clID() const
{ return DeSubSubCalorimeter::classID(); }

// ============================================================================
// standard initialization
// ============================================================================
StatusCode DeSubSubCalorimeter::initialize(){
  /// initialize the base class
  StatusCode sc = DetectorElement::initialize();
  if ( sc.isFailure() ) { return sc ; }

  typedef std::vector<std::string> Parameters;
  typedef Parameters::iterator     Iterator;
  Parameters pars( paramNames() );
  /// cell size
  Iterator it = std::find( pars.begin() , pars.end () , std::string("CellSize") );
  if( pars.end() != it ){
    const double value = param<double>( *it );
    setCellSize( value ) ;
    pars.erase( it );
  }else{ 
      return StatusCode::FAILURE ; 
  }
  /// subSubCalo X size
  Iterator itx = std::find( pars.begin() , pars.end () , std::string("XSize") );
  if( pars.end() != itx ){
    const double value = param<double>( *itx );
    setXSize( value ) ;
    pars.erase( itx );
  }else{ 
    return StatusCode::FAILURE ; 
  }
  /// subSubCalo Y size
  Iterator ity = std::find( pars.begin() , pars.end () , std::string("YSize") );
  if( pars.end() != ity ){
    const double value = param<double>( *ity );
    setYSize( value ) ;
    pars.erase( ity );
  }else{ 
    return StatusCode::FAILURE ; 
  }
  /// area Id 
  Iterator itt = std::find( pars.begin() , pars.end () , std::string("Area") );
  if( pars.end() != itt ){
    const int value = param<int>( *itt );
    setArea( value ) ;
    pars.erase( itt );
  }else{ 
      return StatusCode::FAILURE ; 
  }
  Assert ( 0 != geometry() , "DeSubSubcalorimeter: Invalid GeometryInfo" ) ;
  return StatusCode::SUCCESS;
}

// ============================================================================
// standard printout to STL standard stream
// ============================================================================
std::ostream& DeSubSubCalorimeter::printOut( std::ostream& os ) const{
  os << "\tDeSubSubCalorimeter" << " fullname ='"    << name () << std::endl;
  os << "\t\t\tCellSize=" << std::setw(12) << m_cellSize << " Area-ID = " << m_area  
     << " X-center offset : " << xSize()/2. << std::endl;
  return os;
}

// ============================================================================
// standard printout to Gaudi standard stream
// ============================================================================
MsgStream&    DeSubSubCalorimeter::printOut( MsgStream&    os ) const{
  os << "\tDeSubSubCalorimeter"  << " fullname ='"    << name ()     << endmsg   ;
  os << "\t\t\tCellSize=" << std::setw(12) << m_cellSize << " Area-ID = " << m_area  
     << " X-center offset : " << xSize()/2 << endmsg;
  return os;
}

