/*****************************************************************************\
* (c) Copyright 2018-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#include "CondDBDQScanner.h"

#include "GaudiKernel/IAddressCreator.h"
#include "GaudiKernel/IConverter.h"
#include "GaudiKernel/IDetDataSvc.h"
#include "GaudiKernel/IOpaqueAddress.h"
#include "GaudiKernel/ToolFactory.h"

#include "DetDesc/Condition.h"

// ----------------------------------------------------------------------------
// Implementation file for class: CondDBDQScanner
//
// 04/11/2011: Marco Clemencic
// ----------------------------------------------------------------------------
DECLARE_TOOL_FACTORY( CondDBDQScanner )

namespace {
  class EvtTimeMgr {
  public:
    EvtTimeMgr( IDetDataSvc& dds ) : m_dds( dds ) { m_orig_time = m_dds.eventTime(); }
    void set( const Gaudi::Time& time ) { m_dds.setEventTime( time ); }
    ~EvtTimeMgr() { m_dds.setEventTime( m_orig_time ); }

  private:
    IDetDataSvc& m_dds;
    Gaudi::Time  m_orig_time;
  };
} // namespace

CondDBDQScanner::CondDBDQScanner( const std::string& type, const std::string& name, const IInterface* parent )
    : base_class( type, name, parent ) {

  declareProperty( "ConditionPath", m_condPath = "/Conditions/DQ/Flags",
                   "Path in the Conditions Database where to find the Data "
                   "Quality condition." );

  declareProperty( "Converter", m_converterName = "DetectorPersistencySvc",
                   "Service implementing the IConverter interface." );
}

CondDBDQScanner::~CondDBDQScanner() {}

IDQFilter::FlagsType CondDBDQScanner::scan( const Gaudi::Time& since, const Gaudi::Time& until ) const {
  IDQFilter::FlagsType flags;

  const std::string   par[2]  = {( "git:" + m_condPath ), "Flags"};
  const unsigned long ipar[1] = {0};

  SmartIF<IDetDataSvc> dds( detSvc() );
  if ( !dds ) Exception( "Invalid Detector Data Service" );
  EvtTimeMgr timeMgr( *dds );

  Gaudi::Time cursor = since;
  while ( cursor < until ) {
    timeMgr.set( cursor );
    // Retrieve the condition data
    IOpaqueAddress* addr = NULL;
    DataObject*     obj  = NULL;
    Condition*      cond = NULL;
    if ( m_converter->addressCreator()
             ->createAddress( XML_StorageType, Condition::classID(), par, ipar, addr )
             .isFailure() ||                                    //
         m_converter->createObj( addr, obj ).isFailure() ||     //
         m_converter->fillObjRefs( addr, obj ).isFailure() ||   //
         ( cond = dynamic_cast<Condition*>( obj ) ) == NULL ) { // assignment intended
      delete addr;
      delete obj;
      Exception( "Conversion of Condition failed" );
      return flags; // never reached, but helps Coverity
    }

    // Merge the condition map with the collected one.
    const IDQFilter::FlagsType& condFlags = cond->param<IDQFilter::FlagsType>( "map" );
    flags.insert( condFlags.begin(), condFlags.end() );

    // update time point
    cursor = cond->validTill();

    // finally clean up
    delete addr;
    delete cond;
  }

  return flags;
}

StatusCode CondDBDQScanner::initialize() {
  StatusCode sc = base_class::initialize();
  if ( sc.isFailure() ) return sc;

  m_converter = service( m_converterName );
  if ( UNLIKELY( !m_converter.isValid() ) ) {
    error() << "Cannot get the IConverter implementation " << m_converterName << endmsg;
    return StatusCode::FAILURE;
  }

  return sc;
}

StatusCode CondDBDQScanner::finalize() {
  m_converter.reset();
  return base_class::finalize();
}
