/*****************************************************************************\
* (c) Copyright 2000-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#ifndef PACKPARTICLESANDVERTICES_H
#define PACKPARTICLESANDVERTICES_H 1

#include "GaudiAlg/GaudiAlgorithm.h"
#include "GaudiKernel/AlgFactory.h"
#include "GaudiKernel/IDataManagerSvc.h"
#include "GaudiKernel/SmartIF.h"

#include "Event/StandardPacker.h"
#include "Kernel/Particle2LHCbIDs.h"

#include "Relations/Relation1D.h"

#include "Kernel/LHCbID.h"

#include "Event/PackedTrack.h"
#include "Event/PackedProtoParticle.h"
#include "Event/PackedMuonPID.h"
#include "Event/PackedParticle.h"
#include "Event/PackedVertex.h"
#include "Event/PackedRelations.h"
#include "Event/PackedRecVertex.h"
#include "Event/PackedFlavourTag.h"
#include "Event/PackedPartToRelatedInfoRelation.h"

#include "Event/Track.h"
#include "Event/ProtoParticle.h"
#include "Event/MuonPID.h"
#include "Event/Particle.h"
#include "Event/Vertex.h"
#include "Event/FlavourTag.h"
#include "Event/RecVertex.h"
#include "Event/MCParticle.h"
#include "Event/RelatedInfoMap.h"

/** @class PackParticlesAndVertices PackParticlesAndVertices.h
 *
 *  Packs Particles, Vertices and related information.
 *
 *  @author Olivier Callot
 *  @date   2012-01-23
 */
class PackParticlesAndVertices : public GaudiAlgorithm
{

public:

  /// Standard constructor
  PackParticlesAndVertices( const std::string& name, ISvcLocator* pSvcLocator );

  virtual ~PackParticlesAndVertices( ); ///< Destructor

  virtual StatusCode execute(); ///< Algorithm execution

private:

  typedef LHCb::Relation1D<LHCb::Particle,LHCb::VertexBase>     P2VRELATION;
  typedef LHCb::Relation1D<LHCb::Particle,LHCb::MCParticle>     P2MCPRELATION;
  typedef LHCb::Relation1D<LHCb::Particle,int>                  Part2IntRelations;
  typedef LHCb::Relation1D<LHCb::Particle,LHCb::RelatedInfoMap> Part2InfoRelations; 

private:

  /// Select the data containers in the TES, of a given type, to write out
  void selectContainers ( DataObject* obj,
                          std::vector<std::string>& names,
                          const unsigned int classID,
                          const bool forceRead = false );

  /// Pack a Particle container
  void packAParticleContainer ( const LHCb::Particles* parts,
                                LHCb::PackedParticles& pparts );

  /// Pack a ProtoParticle container
  void packAProtoParticleContainer ( const LHCb::ProtoParticles* protos,
                                     LHCb::PackedProtoParticles& pprotos );

  /// Pack a Muon container
  void packAMuonPIDContainer ( const LHCb::MuonPIDs* pids,
                               LHCb::PackedMuonPIDs& ppids );

  /// Pack a Track container
  void packATrackContainer ( const LHCb::Tracks* tracks,
                             LHCb::PackedTracks& ptracks );

  /// pack a vertex container
  void packAVertexContainer ( const LHCb::Vertices* verts,
                              LHCb::PackedVertices& pverts );

  /// Pack a FlavourTag container
  void packAFTContainer ( const LHCb::FlavourTags* fts,
                          LHCb::PackedFlavourTags& pfts );

  /// Pack a RecVertex container
  void packARecVertexContainer ( const LHCb::RecVertices* rverts,
                                 LHCb::PackedRecVertices& prverts );

  /// Pack a 'SmartRef to SmartRef' relations container
  template < typename RELATION >
  void packAP2PRelationContainer ( const RELATION* rels,
                                   LHCb::PackedRelations& prels );

  /// Pack a 'SmartRef to int' relations container
  template < typename RELATION >
  void packAP2IntRelationContainer ( const RELATION* rels,
                                     LHCb::PackedRelations& prels );

  /// Pack a 'SmartRef to RelatedInfoMap' relations container
  void packAP2RelatedInfoRelationContainer( const Part2InfoRelations * rels,
                                            LHCb::PackedRelatedInfoRelations& prels,
                                            const std::string & location );

private:

  std::string m_inputStream; ///< Input stream root
  bool m_alwaysOutput;      ///< Flag to turn on the creation of output, even when input is missing
  bool m_forceReading;      ///< If true, read the containers from file. Else (def) take what is in TES
  bool m_deleteInput;       ///< delete the containers after packing if true.
  bool m_enableCheck;       ///< Flag to turn on automatic unpacking and checking of the output post-packing
  bool m_listRemaining;     ///< list the remaining objects after packing
  std::vector<std::string> m_vetoedConts; ///< Vetoed containers. Will not be packed.
  StandardPacker m_pack;    ///< Standard packer

};

//=========================================================================
//  Pack a 'SmartRef to SmartRef' relations container
//=========================================================================
template < typename RELATION >
inline void
PackParticlesAndVertices::packAP2PRelationContainer ( const RELATION* rels,
                                                      LHCb::PackedRelations& prels )
{
  // Make a new packed data object and save
  prels.relations().push_back( LHCb::PackedRelation() );
  LHCb::PackedRelation& prel = prels.relations().back();
  // reference to original container and key
  prel.container = m_pack.reference64( &prels, rels, 0 );
  prel.start     = prels.sources().size();
  typename RELATION::Range all = rels->relations();
  for ( typename RELATION::Range::iterator itR = all.begin(); all.end() != itR; ++itR )
  {
    prels.sources().push_back( m_pack.reference64( &prels,
                                                   (*itR).from()->parent(),
                                                   (*itR).from()->key() ) );
    prels.dests().push_back  ( m_pack.reference64( &prels,
                                                   (*itR).to()->parent(),
                                                   (*itR).to()->key()   ) );
  }
  prel.end = prels.sources().size();
  if ( !m_deleteInput ) rels->registry()->setAddress( 0 );
}

//=========================================================================
//  Pack a 'SmartRef to int' relations container
//=========================================================================
template < typename RELATION >
inline void
PackParticlesAndVertices::packAP2IntRelationContainer ( const RELATION* rels,
                                                        LHCb::PackedRelations& prels )
{
  // Make a new packed data object and save
  prels.relations().push_back( LHCb::PackedRelation() );
  LHCb::PackedRelation& prel = prels.relations().back();

  // reference to original container and key
  prel.container = m_pack.reference64( &prels, rels, 0 );

  // First object
  prel.start = prels.sources().size();

  // Loop over relations
  typename RELATION::Range all = rels->relations();
  for ( typename RELATION::Range::iterator itR = all.begin(); all.end() != itR; ++itR )
  {
    prels.sources().push_back( m_pack.reference64( &prels,
                                                   (*itR).from()->parent(),
                                                   (*itR).from()->key() ) );
    prels.dests().push_back  ( (*itR).to() );
  }

  // last object
  prel.end = prels.sources().size();

  // Clear the registry address of the unpacked container, to prevent reloading
  if ( !m_deleteInput ) rels->registry()->setAddress( 0 );
}

#endif // PACKPARTICLESANDVERTICES_H
