/*****************************************************************************\
* (c) Copyright 2000-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// Include files 
#include <vector>
#include "boost/assign/list_of.hpp"

// from Gaudi
#include "GaudiKernel/AlgFactory.h" 
#include "GaudiAlg/GaudiAlgorithm.h" 
#include "Event/RawEvent.h" 
#include "Event/RawBank.h" 

#include "DAQKernel/DecoderAlgBase.h" 


class HltRoutingBitsFilter : public Decoder::AlgBase {
public: 
  HltRoutingBitsFilter( const std::string& name, ISvcLocator* pSvcLocator );
  virtual ~HltRoutingBitsFilter( ); ///< Destructor
  virtual StatusCode initialize();    ///< Algorithm initialisation
  virtual StatusCode execute   ();    ///< Algorithm execution
private:
  std::vector<unsigned int> m_r,m_v;
  
};



//-----------------------------------------------------------------------------
// Implementation file for class : HltRoutingBitsFilter
//
// 2008-07-29 : Gerhard Raven
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( HltRoutingBitsFilter )

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
HltRoutingBitsFilter::HltRoutingBitsFilter( const std::string& name,
                                        ISvcLocator* pSvcLocator)
: Decoder::AlgBase ( name , pSvcLocator )
{

  declareProperty("VetoMask", m_v = std::vector<unsigned int>(3, 0x0));
  declareProperty("RequireMask", m_r = std::vector<unsigned int>(3, 0xFFFF));
  
  //new for decoders, initialize search path, and then call the base method
  m_rawEventLocations = {LHCb::RawEventLocation::Trigger, LHCb::RawEventLocation::Copied, LHCb::RawEventLocation::Default};
  initRawEventSearch();
  

}
//=============================================================================
// Destructor
//=============================================================================
HltRoutingBitsFilter::~HltRoutingBitsFilter() {
} 

//=============================================================================
// Initialisation
//=============================================================================
StatusCode HltRoutingBitsFilter::initialize() {
  StatusCode sc = Decoder::AlgBase::initialize(); // must be executed first
  if (m_v.size()!=3) {
    return Error("Property VetoMask should contain exactly 3 unsigned integers");
  }
  if (m_r.size()!=3) {
    return Error("Property RequireMask should contain exactly 3 unsigned integers");
  }

  return StatusCode::SUCCESS;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode HltRoutingBitsFilter::execute() {

  
  LHCb::RawEvent* rawEvent = findFirstRawEvent();
  if( ! rawEvent ){
    setFilterPassed(true);
    return Error("No RawEvent found at any location",
                 StatusCode::SUCCESS,0);
  }  

  const std::vector<LHCb::RawBank*>& banks = rawEvent->banks(LHCb::RawBank::HltRoutingBits);
  if (banks.size()!=1) {
    setFilterPassed(true);
    return Error("Unexpected # of HltRoutingBits rawbanks",
                 StatusCode::SUCCESS,0);
  }
  if (banks.front()->size()!=3*sizeof(unsigned int)) {
    setFilterPassed(true);
    return Error("Unexpected HltRoutingBits rawbank size",
                 StatusCode::FAILURE,0);
  }
  const unsigned int *data = banks.front()->data();

  
  bool veto = false;
  bool req  = false;
  for (unsigned i=0;i<3 && !veto; ++i) {
        veto = veto || ( data[i] & m_v[i] );
        req  = req  || ( data[i] & m_r[i] );
  }
  bool accept = ( req & !veto );
  counter("#accept") += accept;
  setFilterPassed( accept );

  return StatusCode::SUCCESS;
}
