/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

// Include files

// from Gaudi
#include "GaudiKernel/AlgFactory.h"
#include "GaudiKernel/GaudiException.h"
#include "GaudiAlg/IGenericTool.h"

// event model
#include "Event/GenFSR.h"
#include "Event/GenCountersFSR.h"
#include "Event/CrossSectionsFSR.h"

//local
#include "GenFSRMerge.h"
#include "FSRNavigator.h"  

//-----------------------------------------------------------------------------       
// Implementation file for class : GenFSRMerge                                       
//                                                                                    
// 2015-06-25 : Davide Fazzini                                                           
//-----------------------------------------------------------------------------       

// Declaration of the Algorithm Factory                                               
DECLARE_ALGORITHM_FACTORY( GenFSRMerge )

//=============================================================================       
// Standard constructor, initializes variables                                        
//=============================================================================       

GenFSRMerge::GenFSRMerge(const std::string& name, ISvcLocator* pSvcLocator)
   :  GaudiAlgorithm( name, pSvcLocator),
      m_fileRecordSvc(NULL),
      m_navigatorTool(NULL),
      m_evtType(0),
      m_genName(""),
      m_method("")
{
  // expect the data to be written at LHCb::GenFSRLocation::Default        
  declareProperty( "FileRecordLocation" , m_fileRecordName    = "/FileRecords"   );
  declareProperty( "FSRName"            , m_FSRName           = "/GenFSR"       );
}

//=============================================================================    
// Initialization                                                                  
//=============================================================================    
StatusCode GenFSRMerge::initialize() 
{  
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first          
  if (sc.isFailure()) return sc;  // error printed already by GaudiAlgorithm     
  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Initialize" << endmsg;

  // get the File Records service              
  m_fileRecordSvc = svc<IDataProviderSvc>("FileRecordDataSvc", true);  

  // prepare navigator tool                    
  m_navigatorTool = tool<IFSRNavigator>("FSRNavigator", "FSRNavigator"); 

  return sc;
}

//=============================================================================       
// Main execution                                                                     
//=============================================================================       
StatusCode GenFSRMerge::execute() 
{
  if (msgLevel(MSG::DEBUG)) debug() << "==> Execute" << endmsg;

  return StatusCode::SUCCESS;
}

//=============================================================================       
//  Finalize                                                                          
//=============================================================================       
StatusCode GenFSRMerge::finalize() 
{

  if (msgLevel(MSG::DEBUG)) debug() << "==> Finalize" << endmsg;
  info() << "========== Merging GenFSR ==========" << endmsg;

  // integrate all FSRs in one go                                                     
  const StatusCode mergeSC = merge();
 
  return (mergeSC && GaudiAlgorithm::finalize()); // must be called after all other actions
}

//=============================================================================
//  merge the FSR data
//=============================================================================
StatusCode GenFSRMerge::merge() {
  // make an inventory of the FileRecord store

  std::string slots[] = {"/PrevPrev", "/Prev", "/Next", "/NextNext", ""};
  const int nSlots = sizeof(slots)/sizeof(slots[0]);
  std::vector<std::string> addresses = m_navigatorTool->navigate( m_fileRecordName, m_FSRName );

  for ( int iSlot = 0; iSlot < nSlots; iSlot++ ) {
    std::string   nameFSR = slots[iSlot] + m_FSRName;
    LHCb::GenFSR* genFSR;
    int           nFSRs = 0;

    for ( std::vector<std::string>::iterator iAddr = addresses.begin(); iAddr != addresses.end(); iAddr++) {
      if ( msgLevel( MSG::DEBUG ) ) debug() << "address: " << *iAddr << endmsg;

      if ( iAddr->find( nameFSR ) != std::string::npos ) {
        genFSR = getIfExists<LHCb::GenFSR>( m_fileRecordSvc, *iAddr );
        nFSRs++;

        if ( m_evtType == 0 )
          m_evtType = genFSR->getSimulationInfo( "evtType", 0 );
        else if ( m_evtType != genFSR->getSimulationInfo( "evtType", 0 ) )
          throw GaudiException( "Found a different Event Type! Merging is not allowed!", "GenMergeFSR::finalize",
                                StatusCode::FAILURE );

        if ( m_genName == "" )
          m_genName = genFSR->getSimulationInfo( "hardGenerator", "" );
        else if ( m_genName != genFSR->getSimulationInfo( "hardGenerator", "" ) )
          throw GaudiException( "Found a different Hard Generator Name! Merging is not allowed!",
                                "GenMergeFSR::finalize", StatusCode::FAILURE );

        if ( m_method == "" )
          m_method = genFSR->getSimulationInfo( "generationMethod", "" );
        else if ( m_method != genFSR->getSimulationInfo( "generationMethod", "" ) )
          if ( msgLevel( MSG::VERBOSE ) ) verbose() << "Found a different generation method" << endmsg;
      }
    }

    if ( nFSRs == 0 ) continue;

    LHCb::GenFSR* genFSRMerged = new LHCb::GenFSR();
    put( m_fileRecordSvc, genFSRMerged, m_fileRecordName + nameFSR );

    genFSRMerged->initializeInfos();

    for ( std::vector<std::string>::iterator iAddr = addresses.begin(); iAddr != addresses.end(); iAddr++ ) {
      if ( iAddr->find( nameFSR ) != std::string::npos ) {
        genFSR = getIfExists<LHCb::GenFSR>( m_fileRecordSvc, *iAddr );

        if ( genFSR == NULL ) {
          if ( msgLevel( MSG::VERBOSE ) ) error() << *iAddr << " not found" << endmsg;
        } else {
          if ( msgLevel( MSG::VERBOSE ) ) verbose() << *iAddr << " found" << endmsg;

          // Fill the informations of the new FSR
          if ( genFSRMerged->getSimulationInfo( "hardGenerator", "" ) == "" ) {
            genFSRMerged->setStringInfos( genFSR->stringInfos() );
            genFSRMerged->setIntInfos( genFSR->intInfos() );
          }
          
          // Update the FSR information and append to TS
          *genFSRMerged += *genFSR;
        }
      }
    }

    // read back genFSR from TS
    LHCb::GenFSR* topGenFSR = getIfExists<LHCb::GenFSR>( m_fileRecordSvc, m_fileRecordName + nameFSR );
    if ( msgLevel( MSG::DEBUG ) ) debug() << *topGenFSR << endmsg;

    // clean up original FSRs
    for ( std::vector<std::string>::iterator iAddr = addresses.begin(); iAddr != addresses.end(); iAddr++ ) {
      // get FSR as keyed object and cleanup the original ones - this only cleans genFSRs
      if ( msgLevel( MSG::DEBUG ) ) debug() << "address in list: " << *iAddr << endmsg;

      if ( iAddr->find( nameFSR ) != std::string::npos ) {
        genFSR = getIfExists<LHCb::GenFSR>( m_fileRecordSvc, *iAddr );

        if ( !genFSR ) {
          if ( msgLevel( MSG::ERROR ) ) error() << *iAddr << " not found" << endmsg;
        } else {
          if ( msgLevel( MSG::VERBOSE ) ) verbose() << *iAddr << " found" << endmsg;

          m_fileRecordSvc->unlinkObject( *iAddr ).ignore();
        }
      } 
    }

    // Reset checking conditions
    m_evtType = 0;
    m_genName = "";
    m_method  = "";
    nFSRs     = 0;
  }

  // make a new inventory of the FileRecord store              
  addresses = m_navigatorTool->navigate( m_fileRecordName, m_FSRName );
  for(std::vector<std::string>::iterator iAddr = addresses.begin(); iAddr != addresses.end(); iAddr++)
    if (msgLevel(MSG::DEBUG)) debug() << "address: " << *iAddr << endmsg;
  
  return StatusCode::SUCCESS;
  
}
