/*****************************************************************************\
* (c) Copyright 2000-2023 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// Include files
// local
#include "GenFSRStat.h"

// from Event
#include "Event/CrossSectionsFSR.h"
#include "Event/GenCountersFSR.h"
#include "Event/GenFSR.h"

// From GaudiKernel
#include "GaudiKernel/Time.h"

// other libraries
#include <stdio.h>
#include <cmath>
#include <fstream>
#include <iostream>
#include <string>

#define script_version "20181113"

//-----------------------------------------------------------------------------
// Implementation file for class : GenFSRStat
//
// 2015-08-03 : Davide Fazzini
//-----------------------------------------------------------------------------

namespace {
  //=============================================================================
  //  Get current date/time, format is YYYY-MM-DD.HH:mm:ss
  //=============================================================================

  // naive function to get the current time
  std::string getCurrentTime() { return Gaudi::Time::current().format( true ); }

  //=============================================================================
  //  Get correct event type description
  //=============================================================================

  const std::string getEvtTypeDesc( int evtType ) {
    std::string path_decFile = System::getEnv( "DECFILESROOT" );
    std::string description;

    if ( path_decFile != "" ) {
      std::string  evtType_str = dynamic_cast<std::ostringstream*>(&(std::ostringstream() << evtType))->str();
      std::string  path_desc   = path_decFile + "/options/" + evtType_str + ".py";
      std::fstream file_desc( path_desc.c_str(), std::fstream::in );
      std::string  line_ref  = "$DECFILESROOT/dkfiles/";
      std::string  line_stop = ".dec\"";
      std::string  line;

      while ( getline( file_desc, line ) ) {
        std::size_t pos_ref = line.find( line_ref );

        if ( pos_ref != std::string::npos ) {
          std::size_t pos_end = line.find( line_stop );
          pos_ref             = pos_ref + line_ref.length();
          int count           = pos_end - pos_ref;
          description         = line.substr( pos_ref, count );
          break;
        }
      }
      file_desc.close();
    }
    return description;
  }

  //=============================================================================
  //  Add global statistics in the html file
  //=============================================================================

  void writeGlobalStat( const LHCb::GenFSR& genFSR, std::ostream& htmlOutput ) {
    longlong countEvtGen = 0, countEvtAcc = 0, countIntGen = 0, countIntAcc = 0;

    LHCb::GenFSR::GeneratorCounter::iterator icounter;
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      if ( key > LHCb::GenCountersFSR::CounterKey::IntAccepted ) continue;
      
      if ( key == LHCb::GenCountersFSR::CounterKey::EvtGenerated )
        countEvtGen = icounter->second.second;
      else if ( key == LHCb::GenCountersFSR::CounterKey::IntGenerated )
        countIntGen = icounter->second.second;
      else if ( key == LHCb::GenCountersFSR::CounterKey::EvtAccepted )
        countEvtAcc = icounter->second.second;
      else if ( key == LHCb::GenCountersFSR::CounterKey::IntAccepted )
        countIntAcc = icounter->second.second;
    }
    
    double fraction_events       = genFSR.getEfficiency( countEvtAcc, countEvtGen );
    double fraction_interactions = genFSR.getEfficiency( countIntAcc, countIntGen );

    htmlOutput << "<p> Number of accepted events/generated events: " << countEvtAcc << "/" << countEvtGen;
    htmlOutput << " = " << fraction_events << "</p>\n";
    htmlOutput << "<p> Number of interactions in accepted events/generated interactions: " << countIntAcc;
    htmlOutput << "/" << countIntGen << " = " << fraction_interactions << "</p>\n";
  }

  //=============================================================================
  //  Add generator level counters in the html file
  //=============================================================================

  void writeGeneratorCounters( const LHCb::GenFSR& genFSR, int count, std::ostream& htmlOutput ) {
    LHCb::CrossSectionsFSR             crossFSR;
    std::map<std::string, std::string> mapCross = crossFSR.getFullNames();

    // write the table options
    htmlOutput << "<tr>\n<td align=center>\n<table border=2>\n<th colspan=" << count;
    htmlOutput << "><div class=firstcell>Counters Interaction</div></th>\n" << std::endl;
    htmlOutput << "<tr>\n";

    LHCb::GenFSR::GeneratorCounter::iterator icounter;
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();

    // write the list of cross-section names  
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      if ( key > LHCb::GenCountersFSR::CounterKey::bAndcAcc ) continue;
      if ( key == LHCb::GenCountersFSR::CounterKey::BeforeFullEvt ||
           key == LHCb::GenCountersFSR::CounterKey::AfterFullEvt )
        continue;
      
      std::string name     = LHCb::CrossSectionsFSR::CrossSectionKeyToString( key );
      std::string fullName = mapCross[name];

      longlong before = genFSR.getDenominator( key );
      longlong after  = genFSR.getGenCounterInfo( key ).second;

      if ( before == 0 || after == 0 ) continue;

      if ( key >= LHCb::GenCountersFSR::CounterKey::OnebGen )
        htmlOutput << "<td><font size=2>" << fullName << " (mb)</font></td>\n";
      else
        htmlOutput << "<td><font size=2>" << fullName << "</font></td>\n";
    }

    htmlOutput << "</tr>\n\n<tr>\n";
    
    // write the list of cross-section names    
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      if ( key > LHCb::GenCountersFSR::CounterKey::bAndcAcc ) continue;
      if ( key == LHCb::GenCountersFSR::CounterKey::BeforeFullEvt ||
           key == LHCb::GenCountersFSR::CounterKey::AfterFullEvt )
        continue;
      
      longlong before = genFSR.getDenominator( key );
      if ( before == 0 ) continue;
      
      double C = ( key >= LHCb::GenCountersFSR::CounterKey::OnebGen
                   ? genFSR.getCrossSectionInfo( LHCb::CrossSectionsFSR::CrossSectionKey::MBCrossSection ).second
                   : 1 );
      
      longlong after =
        ( key == LHCb::GenCountersFSR::CounterKey::EvtGenerated ? genFSR.getGenCounterInfo( key + 1 ).second
          : icounter->second.second );
      
      if ( after == 0 ) continue;
      
      double fraction = genFSR.getEfficiency( after, before, C );
      htmlOutput << "<td>" << fraction << "</td>\n";
    }
    htmlOutput << "</tr>\n\n<tr>\n";
    
    // write the list of cross-section names    
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      if ( key > LHCb::GenCountersFSR::CounterKey::bAndcAcc ) continue;
      if ( key == LHCb::GenCountersFSR::CounterKey::BeforeFullEvt ||
           key == LHCb::GenCountersFSR::CounterKey::AfterFullEvt )
        continue;
      
      longlong before = genFSR.getDenominator( key );
      if ( before == 0 ) continue;
      
      double C = ( key >= LHCb::GenCountersFSR::CounterKey::OnebGen
                   ? genFSR.getCrossSectionInfo( LHCb::CrossSectionsFSR::CrossSectionKey::MBCrossSection ).second
                   : 1 );
      
      longlong after =
        ( key == LHCb::GenCountersFSR::CounterKey::EvtGenerated ? genFSR.getGenCounterInfo( key + 1 ).second
          : icounter->second.second );
      
      if ( after == 0 ) continue;
      
      bool flag = ( key != LHCb::CrossSectionsFSR::CrossSectionKey::MeanNoZeroPUInt &&
                    key != LHCb::CrossSectionsFSR::CrossSectionKey::MeanPUInt &&
                    key != LHCb::CrossSectionsFSR::CrossSectionKey::MeanPUIntAcc );
      
      double error = genFSR.getEfficiencyError( after, before, C, flag );
      htmlOutput << "<td>&plusmn; " << error << "</td>\n";
    }
    
    htmlOutput << "</tr>\n</table>\n</td>\n</tr>" << std::endl;
  }
  
  //=============================================================================
  //  Add generator level cut efficiencies in the xml file
  //=============================================================================
  void writeCutEfficiencies( const LHCb::GenFSR& genFSR, int count, std::ostream& htmlOutput ) {
    LHCb::CrossSectionsFSR             crossFSR;
    std::map<std::string, std::string> mapCross = crossFSR.getFullNames();
    
    htmlOutput << "<tr>\n<td align=center>\n<table border=2>\n<th colspan=" << count;
    htmlOutput << "><div class=firstcell>Generator Efficiencies</div></th>\n";
    htmlOutput << "<tr>\n";

    LHCb::GenFSR::GeneratorCounter::iterator icounter;
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    
    // write the list of cross-section names   
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;

      if ( key != LHCb::GenCountersFSR::CounterKey::AfterFullEvt &&
           key != LHCb::GenCountersFSR::CounterKey::AfterLevelCut &&
           key != LHCb::GenCountersFSR::CounterKey::EvtSignal &&
           key != LHCb::GenCountersFSR::CounterKey::EvtantiSignal &&
           key != LHCb::GenCountersFSR::CounterKey::AfterPCut &&
           key != LHCb::GenCountersFSR::CounterKey::AfterantiPCut )
        continue;
      
      std::string name     = LHCb::CrossSectionsFSR::CrossSectionKeyToString( key );
      std::string fullName = mapCross[name];

      longlong before = genFSR.getDenominator( key );
      if ( before == 0 ) continue;

      longlong after = icounter->second.second;
      if ( after == 0 ) continue;

      htmlOutput << "<td><font size=2>" << fullName << "</font></td>\n";
    }
    htmlOutput << "</tr>\n\n<tr>\n";

    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      if ( key != LHCb::GenCountersFSR::CounterKey::AfterFullEvt &&
           key != LHCb::GenCountersFSR::CounterKey::AfterLevelCut &&
           key != LHCb::GenCountersFSR::CounterKey::EvtSignal &&
           key != LHCb::GenCountersFSR::CounterKey::EvtantiSignal &&
           key != LHCb::GenCountersFSR::CounterKey::AfterPCut &&
           key != LHCb::GenCountersFSR::CounterKey::AfterantiPCut )
        continue;
      
      longlong before = genFSR.getDenominator( key );
      if ( before == 0 ) continue;

      longlong after = icounter->second.second;
      if ( after == 0 ) continue;

      double fraction = genFSR.getEfficiency( after, before );
      htmlOutput << "<td>" << fraction << "</td>\n";
      }
    
    htmlOutput << "</tr>\n\n<tr>\n";

    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      if ( key != LHCb::GenCountersFSR::CounterKey::AfterFullEvt &&
           key != LHCb::GenCountersFSR::CounterKey::AfterLevelCut &&
           key != LHCb::GenCountersFSR::CounterKey::EvtSignal &&
           key != LHCb::GenCountersFSR::CounterKey::EvtantiSignal &&
           key != LHCb::GenCountersFSR::CounterKey::AfterPCut &&
           key != LHCb::GenCountersFSR::CounterKey::AfterantiPCut )
        continue;
      
      longlong before = genFSR.getDenominator( key );
      if ( before == 0 ) continue;

      longlong after = icounter->second.second;
      if ( after == 0 ) continue;

      double error = genFSR.getEfficiencyError( after, before );
      htmlOutput << "<td>&plusmn; " << error << "</td>\n";
    }
    htmlOutput << "</tr>\n</table>\n</td>\n</tr>" << std::endl;
  }

  //=============================================================================
  //  Add generated hadron counters in the html file
  //=============================================================================

  void writeGenHadronCounters( const LHCb::GenFSR& genFSR, int countGen, std::ostream& htmlOutput ) {
    LHCb::CrossSectionsFSR             crossFSR;
    std::map<std::string, std::string> mapCross = crossFSR.getFullNames();

    htmlOutput << "<tr>\n<td align=center>\n<table border=2>\n<th colspan=" << countGen;
    htmlOutput << "><div class=firstcell> Generated Hadron Counters</div></th>\n" << std::endl;
    htmlOutput << "<tr>\n";

    LHCb::GenFSR::GeneratorCounter::iterator icounter;
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    // write the list of cross-section names  
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      // Generated Hadron Fractions
      if ( ( key >= LHCb::GenCountersFSR::CounterKey::B0Gen && key <= LHCb::GenCountersFSR::CounterKey::bbGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::D0Gen && key <= LHCb::GenCountersFSR::CounterKey::ccGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::BGen && key <= LHCb::GenCountersFSR::CounterKey::B2starGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::DGen && key <= LHCb::GenCountersFSR::CounterKey::D2starGen ) ) {
        std::string name     = LHCb::CrossSectionsFSR::CrossSectionKeyToString( key );
        std::string fullName = mapCross[name];
        
        longlong before = genFSR.getDenominator( key );
        if ( before == 0 ) continue;

        longlong after = genFSR.getGenCounterInfo( key ).second;
        if ( after == 0 ) continue;

        htmlOutput << "<td><font size=2>" << fullName << "</font></td>\n";
      }
    }

    htmlOutput << "</tr>\n\n<tr>\n";

    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      // Generated Hadron Fractions
      if ( ( key >= LHCb::GenCountersFSR::CounterKey::B0Gen && key <= LHCb::GenCountersFSR::CounterKey::bbGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::D0Gen && key <= LHCb::GenCountersFSR::CounterKey::ccGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::BGen && key <= LHCb::GenCountersFSR::CounterKey::B2starGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::DGen && key <= LHCb::GenCountersFSR::CounterKey::D2starGen ) ) {
        longlong before = genFSR.getDenominator( key );
        if ( before == 0 ) continue;
        
        longlong after = icounter->second.second;
        if ( after == 0 ) continue;

        double fraction = genFSR.getEfficiency( after, before );
        htmlOutput << "<td>" << fraction << "</td>\n";
      }
    }
    htmlOutput << "</tr>\n\n<tr>\n";

    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      // Generated Hadron Fractions
      if ( ( key >= LHCb::GenCountersFSR::CounterKey::B0Gen && key <= LHCb::GenCountersFSR::CounterKey::bbGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::D0Gen && key <= LHCb::GenCountersFSR::CounterKey::ccGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::BGen && key <= LHCb::GenCountersFSR::CounterKey::B2starGen ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::DGen && key <= LHCb::GenCountersFSR::CounterKey::D2starGen ) ) {
        longlong before = genFSR.getDenominator( key );
        if ( before == 0 ) continue;
        longlong after = icounter->second.second;
        if ( after == 0 ) continue;

        double error = genFSR.getEfficiencyError( after, before );
        htmlOutput << "<td>&plusmn; " << error << "</td>\n";
      }
    }
    htmlOutput << "</tr>\n</table>\n</td>\n</tr>" << std::endl;
  }

  //=============================================================================
  //  Add accepted hadron counters in the html file
  //=============================================================================

  void writeAccHadronCounters( const LHCb::GenFSR& genFSR, int countAcc, std::ostream& htmlOutput ) {
    LHCb::CrossSectionsFSR             crossFSR;
    std::map<std::string, std::string> mapCross = crossFSR.getFullNames();

    htmlOutput << "<tr>\n<td align=center>\n<table border=2>\n<th colspan=" << countAcc;
    htmlOutput << "><div class=firstcell> Accepted Hadron Counters</div></th>\n" << std::endl;
    htmlOutput << "<tr>\n";

    LHCb::GenFSR::GeneratorCounter::iterator icounter;
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    
    // write the list of cross-section names   
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      
      // Accepted Hadron Fractions
      if ( ( key >= LHCb::GenCountersFSR::CounterKey::B0Acc && key <= LHCb::GenCountersFSR::CounterKey::bbAcc ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::D0Acc && key <= LHCb::GenCountersFSR::CounterKey::ccAcc ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::BAcc && key <= LHCb::GenCountersFSR::CounterKey::B2starAcc ) ||
           ( key >= LHCb::GenCountersFSR::CounterKey::DAcc && key <= LHCb::GenCountersFSR::CounterKey::D2starAcc ) ) {
        
        std::string name     = LHCb::CrossSectionsFSR::CrossSectionKeyToString( key );
        std::string fullName = mapCross[name];

        longlong before = genFSR.getDenominator( key );
        if ( before == 0 ) continue;

        longlong after = genFSR.getGenCounterInfo( key ).second;
        if ( after == 0 ) continue;

        htmlOutput << "<td><font size=2>" << fullName << "</font></td>\n";
      }
    }

    htmlOutput << "</tr>\n</table>\n</td>\n</tr>" << std::endl;
  }

  //=============================================================================
  //  Add generator cross-sections in the html file
  //=============================================================================

  void writeGeneratorCrossSections( const LHCb::GenFSR& genFSR, int count, std::ostream& htmlOutput ) {
    LHCb::GenFSR::CrossSection::iterator icross;
    LHCb::GenFSR::CrossSection crossSections = genFSR.getCrossSections();

    htmlOutput << "<tr>\n<td align=center>\n<table border=2>\n<th colspan=" << 10;
    htmlOutput << "><div class=firstcell>Hard Generator Cross-Sections</div></th>\n" << std::endl;
    htmlOutput << "<tr>\n";

    std::vector<std::string> temp_names;

    for ( int irow = 0; irow < int( count / 10 ) + 1; irow++ ) {
      int nColumns = 0;

      // Generator Cross-Sections
      for(icross = crossSections.begin(); icross != crossSections.end(); icross++) {
        std::string name = icross->second.first;
        name.erase( name.size() - 2 );
        
        for ( unsigned int jcross = 0; jcross < temp_names.size(); jcross++ )
          if ( name == temp_names.at( jcross ) ) continue;
        
        temp_names.push_back( name );
        
        double value = icross->second.second;
        
        if ( value == 0 || std::isnan( value ) != 0 ) continue;
        
        htmlOutput << "<td><font size=2>" << name << " (mb)</font></td>\n";
        
        nColumns++;
        if ( nColumns == 10 ) break;
      }
      
      htmlOutput << "</tr>\n\n<tr>\n";
      nColumns = 0;
      
      // Generator Cross-Sections
      for(icross = crossSections.begin(); icross != crossSections.end(); icross++) {
        std::string name = icross->second.first;
        name.erase( name.size() - 2 );
        double value = icross->second.second;
        
        for ( unsigned int icross = 0; icross < temp_names.size(); icross++ )
          if ( name == temp_names.at( icross ) ) continue;
        
        if ( value == 0 || std::isnan( value ) != 0 ) continue;
        
        htmlOutput << "<td>" << value << "</td>\n";

        nColumns++;
        if ( nColumns == 10 ) break;
      }

      htmlOutput << "</tr>\n\n<tr>\n";
      htmlOutput << "<tr class=\"blank_row\"><td colspan=\"" << nColumns << "\"></td></tr>\n";
    }

    htmlOutput << "</tr>\n</table>\n</td>\n</tr>" << std::endl;
  }
} // namespace

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( GenFSRStat )

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
GenFSRStat::GenFSRStat( const std::string& name,
                        ISvcLocator* pSvcLocator)
: GaudiAlgorithm (name, pSvcLocator),
                  m_fileRecordSvc(NULL),
                  m_navigatorTool(NULL)
{
  declareProperty("FileRecordLocation" , m_fileRecordName = "/FileRecords");
  declareProperty("FSRName"            , m_FSRName = "/GenFSR");
  declareProperty("HtmlOutputLocation" , m_htmlOutputLocation = "");
  declareProperty("HtmlOutputName"     , m_htmlOutputName = "GenerationFSRnew.html");
  declareProperty("prodID"             , m_prodID = "");
  declareProperty("appConfigVersion"   , m_appConfigVersion = "");
  declareProperty("appConfigFile"      , m_appConfigFile = "");
  declareProperty("gaussVersion"       , m_gaussVersion = "");
  declareProperty("simCond"            , m_simCond = "");
  declareProperty("dddb"               , m_dddb = "");
}

//=============================================================================
// Initialization
//=============================================================================

StatusCode GenFSRStat::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first

  if ( sc.isFailure() ) return sc; // error printed already by GaudiAlgorithm
  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Initialize" << endmsg;

  // get the File Records service
  m_fileRecordSvc = svc<IDataProviderSvc>("FileRecordDataSvc", true);
  m_navigatorTool = tool<IFSRNavigator>( "FSRNavigator", "FSRNavigator" );

  return sc;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode GenFSRStat::execute() {
  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Execute" << endmsg;

  return StatusCode::SUCCESS;
}

//=============================================================================
//  Finalize
//=============================================================================
StatusCode GenFSRStat::finalize() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Finalize" << endmsg;

  GenFSRStat::printFSR();

  return GaudiAlgorithm::finalize(); // must be called after all other actions
}

//=============================================================================

//=============================================================================
//  Print the GenFSR in a file xml
//=============================================================================

void GenFSRStat::printFSR() {
  if ( msgLevel( MSG::DEBUG ) ) debug() << "write to file: " + m_htmlOutputName << endmsg;

  // make an inventory of the FileRecord store
  std::vector<std::string> addresses = m_navigatorTool->navigate( m_fileRecordName, m_FSRName );

  for(std::vector< std::string >::iterator iAddr=addresses.begin(); iAddr!=addresses.end(); iAddr++) {
    DataObject* obj = NULL;
    std::string genRecordAddress = *iAddr;
    StatusCode  sc  = m_fileRecordSvc->retrieveObject( genRecordAddress, obj );

    if ( !sc.isSuccess() ) {
      error() << "Unable to retrieve object '" << genRecordAddress << "'" << endmsg;
      continue;
    }
    LHCb::GenFSR* genFSR = dynamic_cast<LHCb::GenFSR*>( obj );

    if ( genFSR == NULL ) {
      warning() << "genFSR record was not found" << endmsg;
      if ( msgLevel( MSG::DEBUG ) ) debug() << genRecordAddress << " not found" << endmsg;
    } else {
      int evType = genFSR->getSimulationInfo( "evtType", 0 );
      int njobs  = genFSR->getSimulationInfo( "nJobs", 0 );

      std::string genMethod  = genFSR->getSimulationInfo( "generationMethod", "" );
      std::string genName    = genFSR->getSimulationInfo( "hardGenerator", "" );
      std::string decFiles   = genFSR->getSimulationInfo( "decFiles", "" );
      std::string evType_str = dynamic_cast<std::ostringstream*>(&(std::ostringstream() << evType))->str();
      std::string njobs_str  = dynamic_cast<std::ostringstream*>(&(std::ostringstream() << njobs))->str();
      std::string time       = getCurrentTime();

      int count1 = 0, count2 = 0, count3 = 0, count4 = 0, count5 = 0;

      LHCb::GenFSR::GeneratorCounter::iterator icounter;
      LHCb::GenFSR::GeneratorCounter genCounters = genFSR->getGenCounters();

      // look in the FSR what kind if counters are stored
      for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
        int key = icounter->first;

        if ( key == LHCb::GenCountersFSR::CounterKey::AfterFullEvt ||
             key == LHCb::GenCountersFSR::CounterKey::AfterLevelCut ||
             key == LHCb::GenCountersFSR::CounterKey::EvtSignal ||
             key == LHCb::GenCountersFSR::CounterKey::EvtantiSignal ||
             key == LHCb::GenCountersFSR::CounterKey::AfterPCut ||
             key == LHCb::GenCountersFSR::CounterKey::AfterantiPCut ) {
          int before = genFSR->getDenominator( key );
          int after  = genFSR->getGenCounterInfo( key ).second;
          if ( before == 0 || after == 0 ) continue;
          count1 += 1;
        } else if ( key <= LHCb::GenCountersFSR::CounterKey::bAndcAcc &&
                    key != LHCb::GenCountersFSR::CounterKey::BeforeFullEvt &&
                    key != LHCb::GenCountersFSR::CounterKey::AfterFullEvt ) {
          int before = genFSR->getDenominator( key );
          int after  = genFSR->getGenCounterInfo( key ).second;
          if ( before == 0 || after == 0 ) continue;
          count2 += 1;
        } else if ( ( key >= LHCb::GenCountersFSR::CounterKey::B0Gen &&
                      key <= LHCb::GenCountersFSR::CounterKey::bbGen ) ||
                    ( key >= LHCb::GenCountersFSR::CounterKey::D0Gen &&
                      key <= LHCb::GenCountersFSR::CounterKey::ccGen ) ||
                    ( key >= LHCb::GenCountersFSR::CounterKey::BGen &&
                      key <= LHCb::GenCountersFSR::CounterKey::B2starGen ) ||
                    ( key >= LHCb::GenCountersFSR::CounterKey::DGen &&
                      key <= LHCb::GenCountersFSR::CounterKey::D2starGen ) ) {
          int before = genFSR->getDenominator( key );
          int after  = genFSR->getGenCounterInfo( key ).second;
          if ( before == 0 || after == 0 ) continue;
          count3 += 1;
        } else if ( ( key >= LHCb::GenCountersFSR::CounterKey::B0Acc &&
                      key <= LHCb::GenCountersFSR::CounterKey::bbAcc ) ||
                    ( key >= LHCb::GenCountersFSR::CounterKey::D0Acc &&
                      key <= LHCb::GenCountersFSR::CounterKey::ccAcc ) ||
                    ( key >= LHCb::GenCountersFSR::CounterKey::BAcc &&
                      key <= LHCb::GenCountersFSR::CounterKey::B2starAcc ) ||
                    ( key >= LHCb::GenCountersFSR::CounterKey::DAcc &&
                      key <= LHCb::GenCountersFSR::CounterKey::D2starAcc ) ) {
          int before = genFSR->getDenominator( key );
          int after  = genFSR->getGenCounterInfo( key ).second;
          if ( before == 0 || after == 0 ) continue;
          count4 += 1;
        }
      }

      // look in the FSR if there are any generator cross-section is stored
      LHCb::GenFSR::CrossSection::iterator icross;
      LHCb::GenFSR::CrossSection crossSections = genFSR->getCrossSections();

      for(icross = crossSections.begin(); icross != crossSections.end(); icross++) {
        double value = icross->second.second;
        if ( value == 0 || std::isnan( value ) != 0 ) continue;
        count5 += 1;
      }

      // read the already existing file
      std::fstream readOutput( (m_htmlOutputLocation + m_htmlOutputName).c_str(), std::fstream::in );
      // open the new file
      std::string   htmlTempName = "GenerationFSRTemp.html";
      std::ofstream htmlOutput( (m_htmlOutputLocation + htmlTempName).c_str(), std::fstream::out );

      if ( htmlOutput.is_open() ) {
        if ( msgLevel( MSG::DEBUG ) )
          debug() << "Html output: " + m_htmlOutputLocation + htmlTempName + " created." << endmsg;

        if ( readOutput.is_open() ) {
          if ( msgLevel( MSG::DEBUG ) )
            debug() << "Html input: " + m_htmlOutputLocation + m_htmlOutputName + "opened." << endmsg;

          // hyper-link related to this simulation conditions
          std::string ref_evType =
              "<a href=\"GenerationFSR_" + m_appConfigFile + ".html#" + evType_str + "\">" + evType_str + "</a>";

          // link related to this simulation conditions
          std::string link = "<a name=\"" + evType_str + "\"></a>";

          bool        flag_evType = false;
          std::size_t pos_end, pos_ref;
          std::string line;

          // read the existing file GenerationFSR.html
          while ( getline( readOutput, line ) ) {
            // look for the hyper-link in the current line
            if ( !flag_evType ) pos_ref = line.find( ref_evType );

            if ( pos_ref != std::string::npos && !flag_evType ) flag_evType = true;

            // if there is not already a simulation for this production, here we need to add the hyperlink
            // to the table
            if ( line == "</p>" ) {
              if ( flag_evType )
                htmlOutput << line << std::endl;
              else
                htmlOutput << ref_evType << std::endl << line << std::endl;
            }
            // add the tables if the current line contains the right link or if the end of the file is reached
            else if ( ( line == link && flag_evType == true ) || ( flag_evType == false && line == "</body>" ) ) {
              // There is already a simulation for this production
              // but if we are here it's because we have to overwrite the previous results
              htmlOutput << link << std::endl; // write link

              std::string evtDesc = getEvtTypeDesc( evType );

              // Write the tables
              htmlOutput << "<table>\n<th colspan=\"2\"><font color=\"#0000FF\">" << evType_str << " (" << evtDesc
                         << ")</font> ";
              htmlOutput << "Hard Generator <font color=\"#088A08\">" << genName << "</font> ";
              htmlOutput << "Gauss <font color=\"#088A08\">" + m_gaussVersion + "</font>";
              htmlOutput << "<br> DecFiles <font color=\"#088A08\">" + decFiles + "</font>";
              htmlOutput << "<br> DDDB <font color=\"#FF0000\">" + m_dddb + "</font> - ";
              htmlOutput << "SIMCOND <font color=\"#FF0000\">" + m_simCond + "</font><br>";
              htmlOutput << "APPCONFIG <font color=\"#088A08\">" + m_appConfigVersion + "</font> ";
              htmlOutput << "<font color=\"#FF0000\">" + m_appConfigFile + "</font><br>";
              htmlOutput << time << "</th>\n<tr>\n\n";

              if ( count2 != 0 ) writeGeneratorCounters( *genFSR, count2, htmlOutput );
              if ( count3 != 0 ) writeGenHadronCounters( *genFSR, count3, htmlOutput );
              if ( count4 != 0 ) writeAccHadronCounters( *genFSR, count4, htmlOutput );
              if ( count5 != 0 ) writeGeneratorCrossSections( *genFSR, count5, htmlOutput );
              if ( count1 != 0 ) writeCutEfficiencies( *genFSR, count1, htmlOutput );

              htmlOutput << "</table>\n";

              // Write the warnings
              if ( count2 == 0 )
                htmlOutput << "<p><font color=\"#FF0000\">Warning: No Counters Interaction lines found</font></p>\n";
              if ( count3 == 0 && genMethod == "Inclusive" )
                htmlOutput
                    << "<p><font color=\"#FF0000\">Warning: No Generated Hadron Counters lines found</font></p>\n";
              if ( count4 == 0 )
                htmlOutput
                    << "<p><font color=\"#FF0000\">Warning: No Accepted Hadron Counters lines found</font></p>\n";
              if ( count5 == 0 )
                htmlOutput
                    << "<p><font color=\"#FF0000\">Warning: No Hard Generator Cross-Sections lines found</font></p>\n";
              if ( count1 == 0 )
                htmlOutput << "<p><font color=\"#FF0000\">Warning: No Signal Counters lines found</font></p>\n";

              writeGlobalStat( *genFSR, htmlOutput );

              htmlOutput << "<p> Statistics done (script version " << script_version << ") with " << njobs_str
                         << " jobs from the following ProdID:<font color=\"#0000FF\">" + m_prodID + "</font></p>\n";

              // skip the lines which contain the previous table because they have been just updated
              if ( flag_evType ) getline( readOutput, line );

              pos_end = line.find( "<a name=" );

              while ( pos_end == std::string::npos && line != "</body>" ) {
                getline( readOutput, line );
                pos_end = line.find( "<a name=" );
              }

              htmlOutput << line << std::endl;
            } else
              htmlOutput << line << std::endl;
          }
        }
        // the file does not exist, create it
        else {
          if ( msgLevel( MSG::DEBUG ) )
            debug() << m_htmlOutputLocation + m_htmlOutputName << " does not exist, creating it now." << endmsg;

          // write the web-header
          htmlOutput << "<html>\n<head>\n";
          htmlOutput << "<link rel=\"stylesheet\" "
                        "href=\"http://lhcb-release-area.web.cern.ch/LHCb-release-area/DOC/css/lhcb.css\""
                     << " type=\"text/css\" media=\"screen\">\n";
          htmlOutput << "<link rel=\"stylesheet\" "
                        "href=\"http://lhcb-release-area.web.cern.ch/LHCb-release-area/DOC/gauss/css/css.css\""
                     << " type=\"text/css\" media=\"screen\">\n";
          htmlOutput << "<link rel=\"stylesheet\" href=\"css.css\" type=\"text/css\" media=\"screen\">\n";
          htmlOutput << "<title>Table for Generation</title>\n";
          htmlOutput << "</head>\n";

          htmlOutput << "\n<body>\n";
          htmlOutput << "<p> Event Type:\n<a href=\"GenerationFSR_" + m_appConfigFile + ".html#" << evType_str << "\">"
                     << evType_str << "</a>\n</p>\n";

          // link
          htmlOutput << "<a name=\"" << evType_str << "\"></a>\n" << std::endl;

          std::string evtDesc = getEvtTypeDesc( evType );
          // Write the tables
          htmlOutput << "<table>\n<th colspan=\"2\"><font color=\"#0000FF\">" << evType_str << " (" << evtDesc
                     << ")</font> ";
          htmlOutput << "Hard Generator <font color=\"#088A08\">" << genName << "</font> ";
          htmlOutput << "Gauss <font color=\"#088A08\">" + m_gaussVersion + ": ";
          htmlOutput << "</font>";
          htmlOutput << "<br> DecFiles <font color=\"#088A08\">" << decFiles << "</font>";
          htmlOutput << "<br> DDDB <font color=\"#FF0000\">" + m_dddb + "</font> - ";
          htmlOutput << "SIMCOND <font color=\"#FF0000\">" + m_simCond + "</font><br>";
          htmlOutput << "APPCONFIG <font color=\"#088A08\">" + m_appConfigVersion + "</font>";
          htmlOutput << "<font color=\"#FF0000\">" + m_appConfigFile + "</font><br>";
          htmlOutput << time << "</th>\n<tr>\n\n";

          if ( count2 != 0 ) writeGeneratorCounters( *genFSR, count2, htmlOutput );
          if ( count3 != 0 ) writeGenHadronCounters( *genFSR, count3, htmlOutput );
          if ( count4 != 0 ) writeAccHadronCounters( *genFSR, count4, htmlOutput );
          if ( count5 != 0 ) writeGeneratorCrossSections( *genFSR, count5, htmlOutput );
          if ( count1 != 0 ) writeCutEfficiencies( *genFSR, count1, htmlOutput );

          htmlOutput << "</table>\n";

          // Write the warnings
          if ( count2 == 0 )
            htmlOutput << "<p><font color=\"#FF0000\">Warning: No Counters Interaction lines found</font></p>\n";
          if ( count3 == 0 && genMethod == "Inclusive" )
            htmlOutput << "<p><font color=\"#FF0000\">Warning: No Generated Hadron Counters lines found</font></p>\n";
          if ( count4 == 0 )
            htmlOutput << "<p><font color=\"#FF0000\">Warning: No Accepted Hadron Counters lines found</font></p>\n";
          if ( count5 == 0 )
            htmlOutput
                << "<p><font color=\"#FF0000\">Warning: No Hard Generator Cross-Sections lines found</font></p>\n";
          if ( count1 == 0 )
            htmlOutput << "<p><font color=\"#FF0000\">Warning: No Signal Counters lines found</font></p>\n";

          writeGlobalStat( *genFSR, htmlOutput );

          htmlOutput << "<p> Statistics done (script version " << script_version << ") with " << njobs_str
                     << " jobs from the following ProdID:<font color=\"#0000FF\">" + m_prodID + "</font></p>\n";

          htmlOutput << "\n</body>\n</html>\n";
        }
        htmlOutput.close();

        // remove the old file
        std::string originalFile = m_htmlOutputLocation + "GenerationFSR_" + m_appConfigFile + ".html";
        remove(originalFile.c_str());
        std::string newFile = m_htmlOutputLocation + m_htmlOutputName;

        // create the new GenerationFSR.html                                                                                                                                                         
        rename(newFile.c_str(),originalFile.c_str());
      } 
      else if ( msgLevel( MSG::DEBUG ) )
        debug() << "The output file was not opened correctly" << endmsg;
    }
  }
}
